/*	
 *	swipe() :: Swipe action using CSS transforms with jQuery position fallback
 *
 *	Copyright by Lazaworx
 *	http://www.lazaworx.com
 *	Author: Laszlo Molnar
 *
 *	Dual licensed under the MIT and GPL licenses.
 *	- http://www.opensource.org/licenses/mit-license.php
 *	- http://www.gnu.org/copyleft/gpl.html
 *
 *	Usage: $(element).swipe( options );
 */
 
;(function($, $window, $document, undefined) {
	'use strict';
	
	$.fn.swipe = function(options) {
		
		options = $.extend({
						onSwipedLeft:		false,			// Swiped left event
						onSwipedRight:		false,			// Swiped right event
						onSmallSwipe:		false,			// Small swipe (< threshold px)
						onSwipeStart:		false,			// Start swiping event
						onClick:			false,			// Clicked
						onFinished:			false,			// Swiped out - can remove
						threshold:			50,				// Small swipe threshold
						noSwipeInside:		'.caption',		// Avoid swipe inside this container
						swipeoutSpeed:		300,			// Swipe out speed in ms
						overThreshold:		0.25			// Oversized pictures must moved beyond (this * screen width) to trigger swipe 
					}, options);
		
		// Getting coordinates of a touch or mouse event
		var	getDim = function(el) {
					return [ el.outerWidth(), el.outerHeight() ];
				},
				
			getDistance = function(diff) {
					return diff[1]?
						Math.sqrt(Math.pow(diff[0], 2) + Math.pow(diff[1], 2))
						:
						Math.abs(diff[0]);
				},
				
			// Getting coordinates of a touch or mouse event	
			getEventCoords = function(e) {
					if (e.touches && e.touches.length === 1) {
						return [ 
							Math.round(e.touches[0].pageX),
							Math.round(e.touches[0].pageY)
						];
					} else if (e.clientX !== null) {
						return [
							Math.round(e.pageX),
							Math.round(e.pageY)
						];
					}
					return null;
				};
			
		return this.each(function() {
				
			var self = $(this),							// target element
				clip = self.parent(),					// clipping window
				ns = self.data('lsw_ns') ||				// namespace 
					('lsw_' + Math.floor(Math.random() * 10000)),
				selfDim,								// item dimensions
				clipDim,								// clip dimensions
				rto,									// window resize update timeout
				startTime,								// start time
				startPos,								// start position of the mouse
				diff,									// move difference
				//pos,									// current position
				//selfPos,								// the element's position when the swipe starts
				matrix,									// the initial transform matrix
				ended = true,							// swipe ended already
				constrainX = true,						// only horizontal move
				timer = null,
				media = self.hasClass('audio') || self.hasClass('video'), 
				
				// Reading transform matrix
				getMatrix = function() {
						var cs = window.getComputedStyle(self[0], null);
						//console.log('Swipe.js: getMatrix = ' + cs['transform']);
						if (!cs['transform'] || cs['transform'] === 'none') {
							return [ 1, 0, 0, 1, 0, 0 ];
						}
						
						var m = cs.transform.match(/-?\d+\.?\d*/g);
						
						for (var i = 0; i < m.length; i++) {
							m[i] = parseFloat(m[i]);
						}
						
						return m;
					},
					
				// Reset to the original state
				resetTransform = function() {
						//console.log('Swipe.js: Reset transform');
						if (matrix) {
							self[0].style['transform'] = 'matrix(' + matrix.join(',') + ')';
						}
					},
					
				// Stopping ongoing transform
				stopTransition = function() {
						//console.log('Swipe.js: Stop transition');
						var cs = window.getComputedStyle(self[0], null);
						self[0].style['transform'] = cs['transform'];
						self[0].style['transition'] = 'none';
					},
									
				// Reading transform X, Y position
				getPosition = function() {
						var m = getMatrix();
						return m? [ m[4], m[5] ] : [ 0, 0 ];
					},
				
				// Setting transform position
				setPosition = function(pos) {
						//console.log('Swipe.js: setPosition [' + pos.join(',') + ']');
						self.css({
								transform: 		'matrix(' + 
													matrix[0] + ',' + matrix[1] + ',' +
													matrix[2] + ',' + matrix[3] + ',' +  
													pos[0] + ',' + pos[1] + 
												')'
							});
					},
				
				// Scrolling to position and calling complete function
				scroll = function(pos, speed) {
						
						self.css({
								transition:		'transform ' + speed + 'ms ease-out'
							});
						
						self.css({
								transform: 		'matrix(' + 
													matrix[0] + ',' + matrix[1] + ',' +
													matrix[2] + ',' + matrix[3] + ',' +  
													Math.round(pos[0]) + ',' + Math.round(pos[1]) + 
												')'
							});
					},
			
				// Scrolling to position and calling complete function
				scrollAndFade = function(pos, speed, doneFn) {
						//console.log('Swipe.js: scrollAndFade ([' + pos.join(',') + '], ' + speed);
						if (typeof doneFn === FUNCTION) {
							self.one('transitionend', function() {
									self.data('scrolling', false);
									doneFn.call(self);
								});
						}
						
						self.css({
								transition:		'transform ' + speed + 'ms ease-out, opacity ' + speed + 'ms linear'
							});
						
						self.css({
								opacity:		0,
								transform: 		'matrix(' + 
													matrix[0] + ',' + matrix[1] + ',' +
													matrix[2] + ',' + matrix[3] + ',' +  
													Math.round(pos[0]) + ',' + Math.round(pos[1]) + 
												')'
							});
					},
			
				// Starting swipe
				swipeStart = function(e) {
						//log(e.type);
						
						// Avoid swiping during trailing move
						if (self.data('scrolling')) {
							//console.log('Swipe.js: Still scrolling');
							return false;
						}
						
						// Button or link or caption or multifinger or right click
						if (e.target.nodeName === 'BUTTON' || 
							e.target.nodeName === 'A' ||
							$(e.target).closest('.caption').length ||
							(e.originalEvent.touches && e.originalEvent.touches.length > 1) ||
							(e.type === 'mousedown' && e.which !== 1)) {
							//console.log('Swipe.js: Not handled by swipe');
							return true;
						}
						
						// Multi finger touch or right click
						//log('Touch start, fingers: ' + (e.originalEvent.touches? e.originalEvent.touches.length : 0));
						// Audio or Video control bar (bottom 60px : safety zone)
						if (e.target.nodeName === 'AUDIO' || 
							(e.target.nodeName === 'VIDEO' && e.target.controls && 
								e.offsetY > ($(e.target).height() - 60))) {
							// ?????? Does it work?
							$(e.target).trigger('click', e);
							//console.log('Swipe.js: Audio or video controls');
							return true;
						}
						
						//console.log('Swipe.js: Start swiping');
						if (typeof options.onSwipeStart === FUNCTION) {
							options.onSwipeStart.call(self[0], e);
						}
						
						// Allowing long tap
						if (e.type !== 'touchstart') {
							e.preventDefault();
						}
						
						matrix = getMatrix();
						ended = false;
						
						stopTransition();
						self.data('scrolling', false);
						startTime = new Date().getTime();
						
						clipDim = clipDim || getDim(clip);
						selfDim = selfDim || getDim(self);
						
						constrainX = clipDim[1] >= selfDim[1];
						
						startPos = getEventCoords(e.originalEvent);
						diff = [0, 0];
						//pos = getPosition();
						
						// Wiring move event for mouse driven devices
						if (e.type === 'mousedown') {
							$document.on('mousemove.' + ns, swipeMove);
							self.on('mouseup.' + ns + ' mouseout.' + ns, swipeEnd);
						}
						
						return e.type === 'touchstart';
					},
				
				// Moving in progress
				swipeMove = function(e) {
						//log(e.type);
						
						if (e.originalEvent.touches && e.originalEvent.touches.length > 1) {
							// multi finger touch
							return true;
						}
						
						e.preventDefault();
						
						// Event coordinates
						var ec = getEventCoords(e.originalEvent);
						
						diff[0] = ec[0] - startPos[0];
						
						if (constrainX) {
							setPosition([ matrix[4] + diff[0], matrix[5] ]);
						} else {
							diff[1] = ec[1] - startPos[1]; 
							setPosition([ matrix[4] + diff[0], matrix[5] + diff[1] ]);
						}
						//console.log('Swipe.js: Move ' + diff[0] + ':' + diff[1]);
						return false;
					},
				
				// Cancel swipe (e.g. moved out of the screen)
				swipeCancel = function(e) {
						//console.log('Swipe.js: Canceling')
						//log(e.type);
						$document.off('mousemove.' + ns);
						ended = true;
						self.on('click.' + ns, noAction);
						
						setTimeout(function() { 
								self.data('scrolling', false);
								self.off('click.' + ns);
							}, 20);
						
						resetTransform();
						
						return false;
					},
				
				// Stopped moving
				swipeEnd = function(e) {
						//log(e.type);
						if (ended || !matrix) {
							// Multiple triggers for ending swipe
							return true;
						}
						
						e.preventDefault();
						ended = true;
						
						var dt = new Date().getTime() - startTime,
							d = diff? getDistance(diff) : 0,
							speed = 1 + options.swipeoutSpeed / dt,
							
							handleFinished = function() {
									//console.log('Swipe.js: handleFinished');
									timer = clearTimeout(timer);
									self.data('scrolling', false);
									
									if (typeof options.onFinished === FUNCTION) {
										options.onFinished.call(self[0], e);
									}
								},
							
							handleClick = function() {
									//console.log('Swipe.js: handleClick');
									self.data('scrolling', false);
									
									if (diff) {
										if (typeof options.onSmallSwipe === FUNCTION) {
											options.onSmallSwipe.call(self[0], e);
										}
									}
									
									if (typeof options.onClick === FUNCTION) {
										options.onClick.call(self[0], e);
									}
								},
							
							handleSwipe = function(right) {
									//console.log('Swipe.js: handleSwipe(' + right + ')');
									self.data('scrolling', true);
									
									if (right) {
										if (typeof options.onSwipedRight === FUNCTION) {
											options.onSwipedRight.call(self[0], e);
										}
									} else {
										if (typeof options.onSwipedLeft === FUNCTION) {
											options.onSwipedLeft.call(self[0], e);
										}
									}
								};
						
						//console.log('Swipe.js: End swipe, diff = [' + diff.join(',') + ']');
						
						// Removing handlers
						if (e.type === 'mouseup' || e.type === 'mouseout') {
							$document.off('mousemove.' + ns);
							self.off('mouseup.' + ns + ' mouseout.' + ns);
						} else if (e.type === 'pointerup') {
							self.off('pointermove.' + ns + ' pointerup.' + ns);
						}
						
						//self.data('taplength', dt);
						
						if (d > options.threshold) {
							var targetPos = [matrix[4] + diff[0] * speed, matrix[5] + diff[1] * speed];
							// Swipe
							if ((clipDim[0] >= selfDim[0]) ||
									(diff[0] > 0) &&
									((targetPos[0] - selfDim[0] / 2) > clipDim[0] * (options.overTreshold - 0.5)) ||
									((targetPos[0] + selfDim[0] / 2) < clipDim[0] * (0.5 - options.overTreshold))
								) {
								scrollAndFade(targetPos, options.swipeoutSpeed, handleFinished);
								// Swiped enough to be handled as swipedLeft or right
								handleSwipe(diff[0] > 0);
							} else {
								// Not yet reached the boundary: no swipe, just scroll
								scroll(targetPos, options.swipeoutSpeed);
							}
						} else {
							// Small swipe = click
							resetTransform();
							handleClick();
						}
						
						// Avoid click event for 20ms
						self.on('click.' + ns, noAction);
						setTimeout(function() { 
								//self.data('scrolling', false);
								self.off('click.' + ns);
							}, 20);
						
						return false;
					},
				
				// Don't intercept 
				noAction = function(e) {
						e.preventDefault();
						return false;
					},
					
				// Allow selecting in the caption
				selectStart = function(e) {
						return $(e.target).closest('.caption').length > 0;
					},
				
				// Tearing off the swipe handler
				removeSwipe = function() {
						ended = true;
						
						setTimeout(function() { 
								self.data('scrolling', false);
							}, 20);
						
						self.removeAttr('draggable');
						self.add($document).off('.' + ns);
					};
				
			// clean up events that may hang around (in theory this should never happen)
			self.add($document).off('.' + ns);
			
			// Storing namespace
			self.data('lsw_ns', ns);
			
			// Need to be updated about the window size
			$window.on('resize.' + ns, function() {
					clearTimeout(rto);
					rto = setTimeout(function() {
							selfDim = getDim(self);
							clipDim = getDim(clip);		
						}, 50);
				});
			
			self.attr('draggable', 'false')
				.on(TOUCH.START + '.' + ns + ' dragstart.' + ns + ' mousedown.' + ns, swipeStart)
				//.on(TOUCH.CANCEL + '.' + ns + ' dragcancel.' + ns + ' mouseout.' + ns, swipeCancel)
				.on(TOUCH.MOVE + '.' + ns + ' drag.' + ns, swipeMove)
				.on(TOUCH.END + '.' + ns, swipeEnd)
				// External remove request
				.on('removeSwipe.' + ns, removeSwipe)
				.on('selectstart.' + ns, selectStart);
				
			//self.logEvents();
	
		});
	};
			
})(jQuery, jQuery(window), jQuery(document));
